@@ -30,6 +30,7 @@
 #include "nautilus-desktop-directory.h"
 #include "nautilus-global-preferences.h"
 
+#include <eel/eel-debug.h>
 #include <eel/eel-gtk-macros.h>
 #include <eel/eel-glib-extensions.h>
 #include <eel/eel-vfs-extensions.h>
@@ -48,8 +49,6 @@ struct NautilusDesktopLinkMonitorDetails {
 	NautilusDesktopLink *trash_link;
 
 	GList *volume_links;
-
-	GList *mount_black_list;
 };
 
 
@@ -62,12 +61,21 @@ EEL_CLASS_BOILERPLATE (NautilusDesktopLinkMonitor,
 		       G_TYPE_OBJECT)
 
 static NautilusDesktopLinkMonitor *link_monitor = NULL;
-     
+
+static void
+destroy_desktop_link_monitor (void)
+{
+	if (link_monitor != NULL) {
+		g_object_unref (link_monitor);
+	}
+}
+
 NautilusDesktopLinkMonitor *
 nautilus_desktop_link_monitor_get (void)
 {
 	if (link_monitor == NULL) {
 		link_monitor = NAUTILUS_DESKTOP_LINK_MONITOR (g_object_new (NAUTILUS_TYPE_DESKTOP_LINK_MONITOR, NULL));
+		eel_debug_call_at_shutdown (destroy_desktop_link_monitor);
 	}
 	return link_monitor;
 }
@@ -101,23 +109,6 @@ nautilus_desktop_link_monitor_delete_link (NautilusDesktopLinkMonitor *monitor,
 }
 
 
-static gboolean
-volume_in_black_list (NautilusDesktopLinkMonitor *monitor,
-		      const NautilusVolume *volume)
-{
-	GList *p;
-	
-	g_return_val_if_fail (NAUTILUS_IS_DESKTOP_LINK_MONITOR (monitor), TRUE);
-
-	for (p = monitor->details->mount_black_list; p != NULL; p = p->next) {
-		if (strcmp ((char *) p->data, nautilus_volume_get_mount_path (volume)) == 0) {
-			return TRUE;
-		}
-	}
-
-	return FALSE;
-}
-
 static gboolean
 volume_name_exists (NautilusDesktopLinkMonitor *monitor,
 		    const char *name)
@@ -145,7 +136,7 @@ create_volume_link (NautilusDesktopLinkMonitor *monitor,
 	char *unique_name;
 	int index;
 	
-	if (volume_in_black_list (monitor, volume)) {
+	if (nautilus_volume_is_in_removable_blacklist (volume)) {
 		return;
 	}
 
@@ -158,8 +149,9 @@ create_volume_link (NautilusDesktopLinkMonitor *monitor,
 	index = 1;
 
 	unique_name = g_strdup (volume_name);
-	while (volume_name_exists (monitor, volume_name)) {
+	while (volume_name_exists (monitor, unique_name)) {
 		g_free (unique_name);
+		index++;
 		unique_name = g_strdup_printf ("%s (%d)", volume_name, index);
 	}
 
@@ -262,17 +254,11 @@ static void
 nautilus_desktop_link_monitor_init (gpointer object, gpointer klass)
 {
 	NautilusDesktopLinkMonitor *monitor;
-	GList *list;
 
 	monitor = NAUTILUS_DESKTOP_LINK_MONITOR (object);
 
 	monitor->details = g_new0 (NautilusDesktopLinkMonitorDetails, 1);
 
-	/* Set up default mount black list */
-	list = g_list_prepend (NULL, g_strdup ("/proc"));
-	list = g_list_prepend (list, g_strdup ("/boot"));
-	monitor->details->mount_black_list = list;
-
 	/* We keep around a ref to the desktop dir */
 	monitor->details->desktop_dir = nautilus_directory_get (EEL_DESKTOP_URI);
 
@@ -316,7 +302,7 @@ desktop_link_monitor_finalize (GObject *object)
 	}
 
 	if (monitor->details->trash_link != NULL) {
-		g_object_unref (monitor->details->home_link);
+		g_object_unref (monitor->details->trash_link);
 		monitor->details->trash_link = NULL;
 	}
 
@@ -327,9 +313,6 @@ desktop_link_monitor_finalize (GObject *object)
 	nautilus_directory_unref (monitor->details->desktop_dir);
 	monitor->details->desktop_dir = NULL;
 
-	eel_g_list_free_deep (monitor->details->mount_black_list);
-	monitor->details->mount_black_list = NULL;
-
 	eel_preferences_remove_callback (NAUTILUS_PREFERENCES_DESKTOP_HOME_VISIBLE,
 					 desktop_home_visible_changed,
 					 monitor);
